/* Today: a Sine Generator with the ATTiny 25/45/85
 * using Timer 0 in PWM Mode 
 * initially designed to drive a low power air pump for my fish pond
 * from a solar panel
 * the membrane pump comsumes about 4 Watts from 6 Volts AC 
 *
 * A sine table provides the PWM Values and a dead time is inserted for driving 
 * larger loads via high and low side drivers
 * due to pin restrictions we cannot have a lot of user interfacing
 * and we have to set a few parameters in the EEPROM instead
 * fuse the Tiny to 8 Mhz internal RC oscillator and no CLKDIV, BOD Level is very dependent
 * on your powersupply. For my solar panel supply i set it to 1,8 Volts to avoid EEPROM corruption. 
 * 
 * Matthias Schldgen , March 2013
 *
 * by grounding PB2 a simple external frequency control is activated.
 * connect the wiper of a pot to ADC channel 2 (pin 3 on the Tiny = PB4) 
 * and GND & Vcc to the outer ends. 
 * The amplitude of the sine Wave is always set by an analogue voltage
 * on ADC channel 3 ( pin 2 = PB 3). 
 * 
 */

/* USER ADJUSTABLE VALUES 
 * You are free to alter those values to better match your application 
 */

/* The frequency setting for the EEPROM. This value is used when using internal control.
 * E.g. a setting of 210 results in an output frequency of approx. 50 Hz
 * the formula is :
 * (F_CPU/512/256)/(SIN_TABLE_LENGTH/INITIAL_FREQUENCY) = f, where SIN_TABLE_LENGTH is 256
 * and F_CPU typically is 8000000 Hz
 * the maximum frequency with a value of 255 is 61,51 Hz
 * to reach higher frequencies you could shorten the table, where e.g. a SINE_TABLE_LENGTH
 * of 128 would result in a maximum frequency of 123,02 Hz,
 * but the granularity of the sine wave will suffer accordingly
 */
#define INITIAL_FREQUENCY 211
/* set this to one to use inverted outputs for e.g. PNP output stages */
#define USE_INVERTED_OUTPUT 1
/* END OF USER ADJUSTABLE STUFF
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "VFD_table.h"

//Typedefs
/*! \brief Collection of all motor control flags.
 *
 *  This struct contains all motor control flags used in this implementation.
 */
typedef struct PMSMflags
{
  uint8_t externalControl: 1; //! if pot control is selected, this is true
  uint8_t driveWaveform : 2;    //! The current waveform that should be produced.
} PMSMflags_t;

//! FALSE constant value.

#define FALSE   0

//! TRUE constant value, defined to be compatible with comparisons.
#define TRUE    (!FALSE)

//! Bit pattern of PWM pins placed on PORTB.
#define PWM_PATTERN_PORTB   ((1 << PB1) | (1 << PB0))


// jumper this to ground to adjust frequency by external Pot on ADC2 
// Pin 3 (PB 4) on the Tiny 
#define EXTERNAL_CONTROL_PIN  PB2
// connect the freq. pot to this channel
#define ADC_FREQUENCY_PIN   PB4
// amplitude control pot is always active
#define ADC_AMPLITUDE_PIN   PB3

/* How often we update the Speed controller in PWM cycles */
#define CONTROLLER_TIME_BASE 200
/*! The minimum deadtime is used as a fallback if reading the EEPROM fails for any reason
 *  \todo Adjust minimum dead time
 */
#define MIN_DEAD_TIME 3
//! Waveform status flag for sinusoidal driving.
#define WAVEFORM_SINUSOIDAL           1

//! Waveform status flag used in transitions between different types of driving.
#define WAVEFORM_UNDEFINED            3

// definitions for the ADC
// more complicated than necessary as it was stolen from the BLDC controller
//! The ADC channel where the analog frequency control is connected.
#define ADC_CHANNEL_FREQUENCY   2
//! The ADC channel where the analog amplitude input is connected.
#define ADC_CHANNEL_AMPLITUDE    3

//! ADC clock prescaler 2 setting.
#define ADC_PRESCALER_2        ((0 << ADPS2) | (0 << ADPS1) | (1 << ADPS0))

//! ADC clock prescaler 4 setting.
#define ADC_PRESCALER_4        ((0 << ADPS2) | (1 << ADPS1) | (0 << ADPS0))

//! ADC clock prescaler 8 setting.
#define ADC_PRESCALER_8        ((0 << ADPS2) | (1 << ADPS1) | (1 << ADPS0))

//! ADC clock prescaler 16 setting.
#define ADC_PRESCALER_16        ((1 << ADPS2) | (0 << ADPS1) | (0 << ADPS0))

//! ADC clock prescaler 64 setting.
#define ADC_PRESCALER_64        ((1 << ADPS2) | (1 << ADPS1) | (0 << ADPS0))

//! ADC clock prescaler used in this application.
#define ADC_PRESCALER           ADC_PRESCALER_64

//! ADC internal 2.56 volts voltage reference channel value.
#define ADC_REFERENCE_VOLTAGE_25INTERNAL    ((1 << REFS2) |(1 << REFS1) | (0 << REFS0))

//! ADC internal 1.1 bandgap voltage reference channel value.
#define ADC_REFERENCE_VOLTAGE_11INTERNAL    ((0 << REFS2) |(1 << REFS1) | (0 << REFS0))

//! ADC VCC voltage reference channel value.
#define ADC_REFERENCE_VOLTAGE_VCC           ((0 << REFS2) |(0 << REFS1) | (0 << REFS0))

//! ADC AREF voltage reference channel value.
#define ADC_REFERENCE_VOLTAGE_AREF          ((0 << REFS2) |(0 << REFS1) | (1 << REFS0))

/*! ADC voltage reference used in this application.
 *
 *  \todo Select ADC voltage reference channel.
 */
#define ADC_REFERENCE_VOLTAGE   ADC_REFERENCE_VOLTAGE_VCC

//! ADMUX settings for analog frequency measurement.
#define ADMUX_FREQUENCY  (ADC_REFERENCE_VOLTAGE | (1 << ADLAR) | (ADC_CHANNEL_FREQUENCY << MUX0))

//! ADMUX settings for amplitude measurement.
#define ADMUX_AMPLITUDE  (ADC_REFERENCE_VOLTAGE | (1 << ADLAR) | (ADC_CHANNEL_AMPLITUDE << MUX0))


/* Variables */
// Inco determines the output modulator frequency
 /* the formula is :
  * (F_CPU/512/256)/(SIN_TABLE_LENGTH/Inco)
  * we have a sin table length of 256 and a F_CPU of 8 Mhz
  * typical values are 211 for 50 Hz and 252 for 60 Hz
*/
uint8_t ee_Inco __attribute__((section(".eeprom"))) = INITIAL_FREQUENCY;

/*! \brief The frequency of the generated sine waves.
 *
 *  This variable controls the frequency of the generated sine waves.
 *  The range is 0-255.
 */
volatile uint8_t tempU;

volatile uint8_t speedInput;

/*! \brief The amplitude of the generated sine waves.
 *
 *  This variable controls the amplitude of the generated sine waves.
 *  The range is 0-255.
 */
register volatile uint8_t amplitude asm("r6");
// measured amplitude from ADC
volatile uint8_t amplitudeInput;

/*! \brief Flags for Waveform and Direction 
 *
 *  This variable contains all the flags used for VF control.
 */
register volatile PMSMflags_t fastFlags asm ("r9"); 
/*! \brief Increment used for sine table iteration.
 *
 *  This variable holds the increment used for sine table iteration.
 *  It is stored in an 8.8 fixed point format.
 */
register volatile uint16_t sineTableIncrement asm ("r2");

/*! \brief Index into sine table
 *
 *  This variable is used as an index into the sine table. It is
 *  stored in a 8.8 fixed point format, so it can be directly incremented by
 *  sineTableIncrement. Only the high byte is used as table index.
 */
volatile uint16_t sineTableIndex;  
// This is the main frequency control 
register volatile uint8_t Inco asm("r8");

/* start of program */

// more or less only enables the pullups on input pins
// PWM outputs are managed elsewhere
void PortsInit(void) {
 DDRB = 0; // thats right, all inputs here
 PORTB = (1 << EXTERNAL_CONTROL_PIN);
}
/*! \brief Enables PWM output pins
 *
 *  This function enables PWM outputs by setting the port direction for
 *  all PWM pins as output. The PWM configuration itself is not altered
 *  in any way by running this function.
 */
static void EnablePWMOutputs(void)
{
  DDRB |= PWM_PATTERN_PORTB;
}
/*! \brief Disables PWM output pins
 *
 *  This function disables PWM outputs by setting the port direction for
 *  all PWM pins as inputs, thus overriding the PWM. The PWM configuration
 *  itself is not altered in any way by running this function.
 */
static void DisablePWMOutputs(void)
{
  DDRB &= ~PWM_PATTERN_PORTB;
}
/*! \brief Waits for the start of the next PWM cycle.
 *
 *  Waits for the start of the next PWM cycle.
 *  Can be used to make sure that a shoot-through
 *  does not occur in the transition between two output waveform generation modes.
 */
static void TimersWaitForNextPWMCycle(void)
{
  //Clear Timer0 Overflow event flag.
  TIFR = (1 << TOV0);

  //Wait for new event flag.
  while ( !(TIFR & (1 << TOV0)) )
  {}
}
/* init timer to phase correct PWM Mode
 * we run with full speed resulting in a PWM frequency of roughly 15kHz
 */
void TimersInit(void) {

  TCCR0A = (1 << WGM00);
  TCNT0 = 0;
  TCCR0B = (0 << CS01) | (1 << CS00);
}
// now lets initialize the EEPROM and read in our variables
// base frequency
static void EEPROMInit(void)
{
	Inco = eeprom_read_byte(&ee_Inco);
}
/*! 
 *  This function initializes the ADC for frequency and amplitude control
 *  The ADC is triggered once and then relies on 
 *  the ADC complete interrupt service routine to get restarts
 */
static void ADCInit(void)
{
  //Initialize ADC, enable first
  ADCSRA = (1 << ADEN) | (0 << ADSC) | (0 << ADATE) | (1 << ADIF) | (0 << ADIE) | (ADC_PRESCALER);
  // disable digital inputs on analog channels
  DIDR0 = (1 << ADC_FREQUENCY_PIN) | (1 << ADC_AMPLITUDE_PIN); 
 //Select initial AD conversion channel.
  ADMUX = ADMUX_AMPLITUDE;
  //Set trigger source to free running 
  ADCSRB = (0 << ADTS2) | (0 << ADTS1) | (0 << ADTS0);
  //Initialize ADC and start once, no IRQ 
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (0 << ADATE) | (0 << ADIF) | (0 << ADIE) | (ADC_PRESCALER);
 _delay_ms(200);
  //Initialize ADC and restart, now with IRQ 
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (0 << ADATE) | (1 << ADIF) | (1 << ADIE) | (ADC_PRESCALER); 
}
/* select control method by examining the EXTERNAL_CONTROL_PIN
 * if it is jumpered to ground, we derive frequency from analog input
 */
static void DesiredControlUpdate(void)
{
 if ( bit_is_clear(PINB,EXTERNAL_CONTROL_PIN) )
  {
    fastFlags.externalControl = TRUE;
  }
  else
  {
    fastFlags.externalControl = FALSE;
  }
}
/* main program */
int main(void){
 cli();
// Ensure to disable High Speed Mode as our power supply might easily drop to low voltages
 PLLCSR = (1 << LSM);
 PortsInit();
 ADCInit();
 // preset our variables with some values
 Inco=211;
 amplitude=255;
 {
    PMSMflags_t fastFlagsInitial;
    fastFlagsInitial.externalControl = FALSE;
    fastFlagsInitial.driveWaveform = WAVEFORM_UNDEFINED;
    fastFlags = fastFlagsInitial;
  }
// read working values from EEPROM
 EEPROMInit();
// check if we should operate in external control mode
 DesiredControlUpdate();
/*************************** MAIN LOOP   *************************/
 //Enable interrupts globally. The ADC should now run
  sei();
  _delay_ms(100);			// let the ADC sample a few values
  amplitude = amplitudeInput;
// if we run on external contol, we should preset frequency 
  if (fastFlags.externalControl) {
	Inco = speedInput;
	}
  TimersInit();
// enable timer 0 ovf
  TIMSK |= (1 << TOIE0);
/* the main loop is empty as everything happens in the Interrupts */
  for(;;)
  {
   }
}
/* Main End */
/* the following routines are only called from interrupts */

// call the frequency controller
void SpeedController(void) {
if (fastFlags.externalControl) {
  		Inco = speedInput;
   }
  amplitude = amplitudeInput;
}
// set timer to phase correct PWM and output it on OC0A and OC0B
static void TimersSetModeSinusoidal(void)
{
  //Set PWM pins to input (Hi-Z) while changing modes.
DisablePWMOutputs();
#if (USE_INVERTED_OUTPUT)
  // initialize inverted pair output
  TCCR0A = (1 << COM0A1) | (1 << COM0A0) | (1 << COM0B1) | (1 << COM0B0) | (1 << WGM00);
// makes sure outputs are off 
  OCR0B = 0xff;
  OCR0A = 0xff;
#else
  // initialize normal pair output
  TCCR0A = (1 << COM0A1) | (0 << COM0A0) | (1 << COM0B1) | (0 << COM0B0) | (1 << WGM00);
// makes sure outputs are off 
  OCR0A = 0;
  OCR0B = 0;
#endif
  //Wait for next PWM cycle to ensure that all outputs are updated.
  TimersWaitForNextPWMCycle();

  fastFlags.driveWaveform = WAVEFORM_SINUSOIDAL;
 //Change PWM pins to output again to allow PWM control.
  EnablePWMOutputs();
}

/*! \brief Adjusts the sine table index according to the current increment.
 *
 *  This function increases the sine table index with the given increment
 *  The index is then adjusted to be within the table length.
 *
 *  \param increment The increment (in 8.8 format) added to the sine table index.
 */
static void AdjustSineTableIndex(const uint16_t increment)
{
  sineTableIndex += increment ;

  // If the table index is out of bounds, wrap the index around the table end
  // to continue from the beginning. 
  if ((sineTableIndex >> 8) >= SINE_TABLE_LENGTH)
  {
    sineTableIndex -= (SINE_TABLE_LENGTH << 8);
  }
}
/* interrupt service routines */
// ISR stub for  unused irqs. The AVRs sometimes need to fire an interupt and 
// execute it before OC registers are updated
EMPTY_INTERRUPT(TIMER0_COMPB_vect);

ISR(TIMER0_COMPA_vect,ISR_ALIASOF (TIMER0_COMPB_vect));


/* The ISRs are the core routines of the frequency generator
* while the ADC ISR reads frequency input and amplitude, the Timer0 OVF ISR does
* the main work for generating sine modulated PWM
*/
/*! \brief Timer1 Capture Event interrupt service routine.
 *
 * This interrupt service routine is run everytime the timer0
 * encounters an overflow. New sinusoidal output value is calculated and the
 * timer is updated to reflect the new values.
 * The core routine of the sine Drive
 */
ISR(TIMER0_OVF_vect) {

static uint8_t speedRegTicks = 0;
int16_t tempScale;
/* preset the table pointer to start of the table */
uint8_t *sineTablePtr = sineTable ;

// disable PWM for frequency = 0
  if (Inco < 1) {
      	fastFlags.driveWaveform = WAVEFORM_UNDEFINED;
	  	DisablePWMOutputs();
	  	return;
	  }
	  else {
  		if (fastFlags.driveWaveform != WAVEFORM_SINUSOIDAL) {
            TimersSetModeSinusoidal() ;  
    	}
  	}
  sineTableIncrement = Inco;
  AdjustSineTableIndex(sineTableIncrement);
//Add sine table offset to pointer 
  sineTablePtr += (sineTableIndex >> 8);
/* now read the working value from table */
  tempU = pgm_read_byte(sineTablePtr);
/* Scale sine modulation values to the desired amplitude 
 * we first remove the 128er offset then multiply by amplitude (resulting in a signed 16 bit)
 * shift the 16 bit into the 8 bit result register and then we add the offset again
 */
	tempScale = (tempU - 128) * amplitude;
	tempU = 128 + (tempScale >> 8); 
/* set the OC registers depending on the halfwave 
 * The upper half (sine value above 128) is modulated by OCR0A, while OCR0B stays low
 * The lower half is done by OCR0B while OCR0A stays low 
 * The H-Bridge thus changes polarity every half sine wave
 */
  if (tempU > 127) {
   	OCR0A = (tempU - 128) << 1;
   	OCR0B = 1;
   } else {
   	OCR0A = 1;
   	OCR0B = ((127 - tempU) << 1);
   };
/* regular task call the SpeedController to read ADC values */
  speedRegTicks++;
    if (speedRegTicks >= CONTROLLER_TIME_BASE)
    {
		SpeedController();
     		speedRegTicks = 0; 
	}

}
/* ADC complete interrupt */
ISR(ADC_vect, ISR_NOBLOCK)
{
  if (fastFlags.externalControl) {
  	switch (ADMUX)
  	{
  		case ADMUX_FREQUENCY:
   		 	speedInput = (speedInput + ADCH) >> 1; // do some averaging
    		ADMUX = ADMUX_AMPLITUDE;
    	break;
  		case ADMUX_AMPLITUDE:
    		amplitudeInput = (amplitudeInput + ADCH) >> 1;
    		ADMUX = ADMUX_FREQUENCY; break;
  		default: break;
   	}
  } else {
   	amplitudeInput = (amplitudeInput + ADCH) >> 1;
    ADMUX = ADMUX_AMPLITUDE;
    }
// restart the ADC
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (0 << ADATE) | (1 << ADIF) | (1 << ADIE) | ADC_PRESCALER;
}
